package ru.dbotthepony.kstarbound.defs.world

import com.google.gson.JsonObject
import ru.dbotthepony.kommons.gson.set
import ru.dbotthepony.kommons.math.RGBAColor
import ru.dbotthepony.kommons.util.AABBi
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.GlobalDefaults
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.fromJson
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.util.random.nextRange
import ru.dbotthepony.kstarbound.util.random.random
import kotlin.properties.Delegates

class AsteroidsWorldParameters : VisitableWorldParameters() {
	var asteroidTopLevel: Int = 0
		private set
	var asteroidBottomLevel: Int = 0
		private set
	var blendSize: Double = 0.0
		private set
	var asteroidBiome: String by Delegates.notNull()
		private set
	var ambientLightLevel: RGBAColor by Delegates.notNull()
		private set

	init {
		airless = true
	}

	override val type: VisitableWorldParametersType
		get() = VisitableWorldParametersType.ASTEROIDS

	@JsonFactory
	data class StoreData(
		val asteroidTopLevel: Int = 0,
		val asteroidBottomLevel: Int = 0,
		val blendSize: Double = 0.0,
		val asteroidBiome: String,
		val ambientLightLevel: RGBAColor,
	)

	override fun fromJson(data: JsonObject) {
		super.fromJson(data)

		val store = Starbound.gson.fromJson(data, StoreData::class.java)
		asteroidTopLevel = store.asteroidTopLevel
		asteroidBottomLevel = store.asteroidBottomLevel
		blendSize = store.blendSize
		asteroidBiome = store.asteroidBiome
		ambientLightLevel = store.ambientLightLevel
	}

	override fun toJson(data: JsonObject, isLegacy: Boolean) {
		super.toJson(data, isLegacy)

		val store = Starbound.gson.toJsonTree(StoreData(
			asteroidTopLevel = asteroidTopLevel,
			asteroidBottomLevel = asteroidBottomLevel,
			blendSize = blendSize,
			asteroidBiome = asteroidBiome,
			ambientLightLevel = ambientLightLevel,
		)) as JsonObject

		for ((k, v) in store.entrySet())
			data[k] = v
	}

	override fun createLayout(seed: Long): WorldLayout {
		val random = random(seed)
		val terrain = GlobalDefaults.asteroidWorlds.terrains.random(random)

		val layout = WorldLayout()
		layout.worldSize = worldSize

		val asteroidRegion = WorldLayout.RegionParameters(
			worldSize.y / 2,
			threatLevel,
			asteroidBiome,
			terrain.terrainSelector,
			terrain.caveSelector,
			terrain.bgCaveSelector,
			terrain.oreSelector,
			terrain.oreSelector,
			terrain.subBlockSelector,
			WorldLayout.RegionLiquids(),
		)

		val emptyRegion = WorldLayout.RegionParameters(
			worldSize.y / 2,
			threatLevel,
			asteroidBiome,
			GlobalDefaults.asteroidWorlds.emptyTerrain.terrainSelector,
			GlobalDefaults.asteroidWorlds.emptyTerrain.caveSelector,
			GlobalDefaults.asteroidWorlds.emptyTerrain.bgCaveSelector,
			GlobalDefaults.asteroidWorlds.emptyTerrain.oreSelector,
			GlobalDefaults.asteroidWorlds.emptyTerrain.oreSelector,
			GlobalDefaults.asteroidWorlds.emptyTerrain.subBlockSelector,
			WorldLayout.RegionLiquids(),
		)

		layout.addLayer(random, 0, emptyRegion)
		layout.addLayer(random, asteroidBottomLevel, asteroidRegion)
		layout.addLayer(random, asteroidTopLevel, emptyRegion)

		layout.regionBlending = blendSize
		layout.blockNoise = GlobalDefaults.asteroidWorlds.blockNoise?.build(random)

		layout.playerStartSearchRegions.add(
			AABBi(
				Vector2i(0, asteroidBottomLevel),
				Vector2i(worldSize.x, asteroidTopLevel),
			)
		)

		layout.finalize(RGBAColor.BLACK)
		return layout
	}

	companion object {
		fun generate(seed: Long): AsteroidsWorldParameters {
			val random = random(seed)

			val parameters = AsteroidsWorldParameters()

			parameters.threatLevel = random.nextRange(GlobalDefaults.asteroidWorlds.threatRange)
			parameters.typeName = "asteroids"
			parameters.worldSize = GlobalDefaults.asteroidWorlds.worldSize
			parameters.gravity = Vector2d(y = random.nextRange(GlobalDefaults.asteroidWorlds.gravityRange))
			parameters.environmentStatusEffects = GlobalDefaults.asteroidWorlds.environmentStatusEffects
			parameters.overrideTech = GlobalDefaults.asteroidWorlds.overrideTech
			parameters.globalDirectives = GlobalDefaults.asteroidWorlds.globalDirectives
			parameters.beamUpRule = GlobalDefaults.asteroidWorlds.beamUpRule
			parameters.disableDeathDrops = GlobalDefaults.asteroidWorlds.disableDeathDrops
			parameters.worldEdgeForceRegions = GlobalDefaults.asteroidWorlds.worldEdgeForceRegions
			parameters.asteroidTopLevel = GlobalDefaults.asteroidWorlds.asteroidsTop
			parameters.asteroidBottomLevel = GlobalDefaults.asteroidWorlds.asteroidsBottom
			parameters.blendSize = GlobalDefaults.asteroidWorlds.blendSize
			parameters.ambientLightLevel = GlobalDefaults.asteroidWorlds.ambientLightLevel
			parameters.asteroidBiome = GlobalDefaults.asteroidWorlds.biome

			return parameters
		}
	}
}
