package ru.dbotthepony.kstarbound.world

import com.google.gson.JsonObject
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap
import it.unimi.dsi.fastutil.ints.IntArraySet
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import it.unimi.dsi.fastutil.objects.ReferenceOpenHashSet
import ru.dbotthepony.kommons.arrays.Object2DArray
import ru.dbotthepony.kommons.collect.filterNotNull
import ru.dbotthepony.kommons.util.IStruct2d
import ru.dbotthepony.kommons.util.IStruct2i
import ru.dbotthepony.kommons.util.AABB
import ru.dbotthepony.kommons.util.MailboxExecutorService
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.world.WorldStructure
import ru.dbotthepony.kstarbound.defs.world.WorldTemplate
import ru.dbotthepony.kstarbound.math.*
import ru.dbotthepony.kstarbound.network.IPacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.SetPlayerStartPacket
import ru.dbotthepony.kstarbound.util.ParallelPerform
import ru.dbotthepony.kstarbound.world.api.ICellAccess
import ru.dbotthepony.kstarbound.world.api.AbstractCell
import ru.dbotthepony.kstarbound.world.api.TileView
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import ru.dbotthepony.kstarbound.world.entities.DynamicEntity
import ru.dbotthepony.kstarbound.world.entities.TileEntity
import ru.dbotthepony.kstarbound.world.physics.CollisionPoly
import ru.dbotthepony.kstarbound.world.physics.CollisionType
import ru.dbotthepony.kstarbound.world.physics.Poly
import ru.dbotthepony.kstarbound.world.physics.getBlockPlatforms
import ru.dbotthepony.kstarbound.world.physics.getBlocksMarchingSquares
import java.io.Closeable
import java.util.concurrent.ForkJoinPool
import java.util.concurrent.atomic.AtomicInteger
import java.util.concurrent.locks.ReentrantLock
import java.util.function.Predicate
import java.util.random.RandomGenerator
import java.util.stream.Stream

abstract class World<This : World<This, ChunkType>, ChunkType : Chunk<This, ChunkType>>(val template: WorldTemplate) : ICellAccess, Closeable {
	val background = TileView.Background(this)
	val foreground = TileView.Foreground(this)
	val mailbox = MailboxExecutorService()
	val sky = Sky()
	val geometry: WorldGeometry = template.geometry

	val nextEntityID = AtomicInteger()

	override fun getCellDirect(x: Int, y: Int): AbstractCell {
		if (!geometry.x.inBoundsCell(x) || !geometry.y.inBoundsCell(y)) return AbstractCell.NULL
		return getCell(x, y)
	}

	override fun getCell(x: Int, y: Int): AbstractCell {
		return chunkMap.getCell(x, y)
	}

	override fun setCell(x: Int, y: Int, cell: AbstractCell): Boolean {
		return chunkMap.setCell(x, y, cell)
	}

	protected open fun onChunkCreated(chunk: ChunkType) { }
	protected open fun onChunkRemoved(chunk: ChunkType) { }

	abstract inner class ChunkMap : Iterable<ChunkType> {
		abstract operator fun get(x: Int, y: Int): ChunkType?
		abstract fun compute(x: Int, y: Int): ChunkType?
		fun compute(pos: ChunkPos) = compute(pos.x, pos.y)

		abstract fun remove(x: Int, y: Int)
		fun remove(pos: ChunkPos) = remove(pos.x, pos.y)

		abstract fun getCell(x: Int, y: Int): AbstractCell
		abstract fun setCell(x: Int, y: Int, cell: AbstractCell): Boolean

		operator fun get(pos: ChunkPos) = get(pos.x, pos.y)

		protected fun create(x: Int, y: Int): ChunkType {
			val pos = ChunkPos(x, y)
			val chunk = chunkFactory(pos)
			val orphanedInThisChunk = ArrayList<AbstractEntity>()

			for (ent in orphanedEntities) {
				if (ent.chunkPos == pos) {
					orphanedInThisChunk.add(ent)
				}
			}

			for (ent in orphanedInThisChunk) {
				ent.chunk = chunk
			}

			return chunk
		}

		abstract val size: Int
	}

	// around 30% slower than ArrayChunkMap, but can support insanely large worlds
	inner class SparseChunkMap : ChunkMap() {
		private val map = Long2ObjectOpenHashMap<ChunkType>()

		override fun getCell(x: Int, y: Int): AbstractCell {
			if (!geometry.x.isValidCellIndex(x) || !geometry.y.isValidCellIndex(y)) return AbstractCell.NULL
			val ix = geometry.x.cell(x)
			val iy = geometry.y.cell(y)
			return this[geometry.x.chunkFromCell(ix), geometry.y.chunkFromCell(iy)]?.getCell(ix and CHUNK_SIZE_MASK, iy and CHUNK_SIZE_MASK) ?: AbstractCell.NULL
		}

		override fun get(x: Int, y: Int): ChunkType? {
			if (!geometry.x.inBoundsChunk(x) || !geometry.y.inBoundsChunk(y)) return null
			return map[ChunkPos.toLong(x, y)]
		}

		override fun compute(x: Int, y: Int): ChunkType? {
			if (!geometry.x.inBoundsChunk(x) || !geometry.y.inBoundsChunk(y)) return null

			val index = ChunkPos.toLong(x, y)
			val get = map[index] ?: create(x, y).also { map[index] = it; onChunkCreated(it) }
			return get
		}

		override fun setCell(x: Int, y: Int, cell: AbstractCell): Boolean {
			if (!geometry.x.isValidCellIndex(x) || !geometry.y.isValidCellIndex(y)) return false
			val ix = geometry.x.cell(x)
			val iy = geometry.y.cell(y)
			val cx = geometry.x.chunkFromCell(ix)
			val cy = geometry.y.chunkFromCell(iy)

			val index = ChunkPos.toLong(cx, cy)

			val get = map[index] ?: create(cx, cy).also { map[index] = it }
			return get.setCell(ix and CHUNK_SIZE_MASK, iy and CHUNK_SIZE_MASK, cell)
		}

		override fun remove(x: Int, y: Int) {
			val index = ChunkPos.toLong(geometry.x.chunk(x), geometry.y.chunk(y))
			val chunk = map.get(index)

			if (chunk != null) {
				chunk.remove()
				onChunkRemoved(chunk)
				map.remove(index)
			}
		}

		override fun iterator(): Iterator<ChunkType> {
			return map.values.iterator()
		}

		override val size: Int
			get() = map.size
	}

	inner class ArrayChunkMap : ChunkMap() {
		private val map = Object2DArray.nulls<ChunkType>(divideUp(geometry.size.x, CHUNK_SIZE), divideUp(geometry.size.y, CHUNK_SIZE))
		private val existing = ObjectArraySet<ChunkPos>()

		private fun getRaw(x: Int, y: Int): ChunkType? {
			return map[x, y]
		}

		override fun compute(x: Int, y: Int): ChunkType? {
			if (!geometry.x.inBoundsChunk(x) || !geometry.y.inBoundsChunk(y)) return null
			return map[x, y] ?: create(x, y).also { existing.add(ChunkPos(x, y)); map[x, y] = it; onChunkCreated(it) }
		}

		override fun getCell(x: Int, y: Int): AbstractCell {
			if (!geometry.x.isValidCellIndex(x) || !geometry.y.isValidCellIndex(y)) return AbstractCell.NULL
			val ix = geometry.x.cell(x)
			val iy = geometry.y.cell(y)
			return map[ix ushr CHUNK_SIZE_BITS, iy ushr CHUNK_SIZE_BITS]?.getCell(ix and CHUNK_SIZE_MASK, iy and CHUNK_SIZE_MASK) ?: AbstractCell.NULL
		}

		override fun setCell(x: Int, y: Int, cell: AbstractCell): Boolean {
			if (!geometry.x.isValidCellIndex(x) || !geometry.y.isValidCellIndex(y)) return false
			val ix = geometry.x.cell(x)
			val iy = geometry.y.cell(y)
			return compute(ix ushr CHUNK_SIZE_BITS, iy ushr CHUNK_SIZE_BITS)!!.setCell(ix and CHUNK_SIZE_MASK, iy and CHUNK_SIZE_MASK, cell)
		}

		override fun get(x: Int, y: Int): ChunkType? {
			if (!geometry.x.inBoundsChunk(x) || !geometry.y.inBoundsChunk(y)) return null
			return map[x, y]
		}

		@Suppress("NAME_SHADOWING")
		override fun remove(x: Int, y: Int) {
			val x = geometry.x.chunk(x)
			val y = geometry.y.chunk(y)
			val chunk = map[x, y]

			if (chunk != null) {
				chunk.remove()
				onChunkRemoved(chunk)
				existing.remove(ChunkPos(x, y))
				map[x, y] = null
			}
		}

		override fun iterator(): Iterator<ChunkType> {
			val parent = existing.iterator()

			return object : Iterator<ChunkType> {
				override fun hasNext(): Boolean {
					return parent.hasNext()
				}

				override fun next(): ChunkType {
					val (x, y) = parent.next()
					return map[x, y] ?: throw ConcurrentModificationException()
				}
			}
		}

		override val size: Int
			get() = existing.size
	}

	val chunkMap: ChunkMap = if (geometry.size.x <= 32000 && geometry.size.y <= 32000) ArrayChunkMap() else SparseChunkMap()

	val random: RandomGenerator = RandomGenerator.of("Xoroshiro128PlusPlus")
	var gravity = Vector2d(0.0, -80.0)
	abstract val isRemote: Boolean

	// generic lock
	val lock = ReentrantLock()

	val orphanedEntities = ReferenceOpenHashSet<AbstractEntity>()
	val entities = Int2ObjectOpenHashMap<AbstractEntity>()
	val dynamicEntities = ReferenceOpenHashSet<DynamicEntity>()
	val tileEntities = ReferenceOpenHashSet<TileEntity>()

	var playerSpawnPosition = Vector2d.ZERO
		protected set

	var respawnInWorld = false
		protected set

	var adjustPlayerSpawn = false

	var ticks = 0L
		private set

	open fun broadcast(packet: IPacket) {

	}

	var centralStructure: WorldStructure = WorldStructure()

	val protectedDungeonIDs = IntArraySet()
	val properties = JsonObject()

	open fun setPlayerSpawn(position: Vector2d, respawnInWorld: Boolean) {
		playerSpawnPosition = position
		this.respawnInWorld = respawnInWorld
		broadcast(SetPlayerStartPacket(position, respawnInWorld))
	}

	abstract fun isSameThread(): Boolean

	fun ensureSameThread() {
		check(isSameThread()) { "Trying to access $this from ${Thread.currentThread()}" }
	}

	fun tick() {
		try {
			ticks++
			mailbox.executeQueuedTasks()

			ForkJoinPool.commonPool().submit(ParallelPerform(dynamicEntities.spliterator(), { if (!it.isRemote) it.movement.move() })).join()
			mailbox.executeQueuedTasks()

			entities.values.forEach { it.think() }
			mailbox.executeQueuedTasks()

			for (chunk in chunkMap) {
				chunk.tick()
			}

			mailbox.executeQueuedTasks()
			tickInner()
		} catch(err: Throwable) {
			throw RuntimeException("Ticking world $this", err)
		}
	}

	protected abstract fun tickInner()
	protected abstract fun chunkFactory(pos: ChunkPos): ChunkType

	override fun close() {
		mailbox.shutdownNow()
	}

	fun queryCollisions(aabb: AABB): MutableList<CollisionPoly> {
		val result = ArrayList<CollisionPoly>()
		val tiles = aabb.encasingIntAABB()

		for (x in tiles.mins.x .. tiles.maxs.x) {
			for (y in tiles.mins.y .. tiles.maxs.y) {
				getBlocksMarchingSquares(x, y, foreground, CollisionType.DYNAMIC, result)
				getBlockPlatforms(x, y, foreground, CollisionType.PLATFORM, result)
			}
		}

		return result
	}

	fun collide(with: Poly, filter: Predicate<CollisionPoly>): Stream<Poly.Penetration> {
		return queryCollisions(with.aabb.enlarge(1.0, 1.0)).stream()
			.filter(filter)
			.map { with.intersect(it.poly) }
			.filterNotNull()
	}

	fun polyIntersects(with: Poly, filter: Predicate<CollisionPoly> = Predicate { true }, tolerance: Double = 0.0): Boolean {
		return collide(with, filter).anyMatch { it.penetration >= tolerance }
	}

	fun polyIntersects(with: Poly, filter: Collection<CollisionType>, tolerance: Double = 0.0): Boolean {
		return polyIntersects(with, Predicate { it.type in filter }, tolerance)
	}

	fun gravityAt(pos: IStruct2i): Vector2d {
		return gravity
	}

	fun gravityAt(pos: IStruct2d): Vector2d {
		return gravity
	}
}
